package com.astrum.io;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;

public class IOGpioThread extends IONative {
	static final String THREAD_NAME = "IOGpioThread";

	public interface IOGpioState {
		public void onGpioStateChange(IOGpioInput gpio, boolean value) throws Exception;

		public void onGpioState(IOGpioInput gpio, boolean value) throws Exception;
	}

	class GPIOValueList extends ArrayList<Boolean> {
		@Override
		public boolean add(Boolean object) {
			if (super.size() > 2) {
				super.remove(super.size() - 1);
			}
			super.add(0, object);
			return true;
		}

		public boolean get() {
			if (this.size() < 2) {
				return true;
			}
			boolean result = this.get(0);
			for (int i = 0; i < this.size(); i++) {
				if (result != this.get(i)) {
					return true;
				}
			}
			return result;
		}
	}

	HashMap<IOGPIO, GPIOValueList> gpioList = new HashMap<>();
	HashMap<String, IOGPIO> gpioMap = new HashMap<>();
	ReentrantLock lock = new ReentrantLock();
	GPIOThread thread = null;
	List<IOGpioState> gpioStateList = new ArrayList<>();

	int requestTimeOut = 200;
	boolean paused = false;

	public boolean isPaused() {
		return paused;
	}

	public void setPaused(boolean paused) {
		this.paused = paused;
	}

	public IOGpioThread(int requestTimeOut) {
		this.requestTimeOut = requestTimeOut;
	}

	public void register(IOGPIO gpio) {
		try {
			lock.lock();
			if (gpio.open()) {
				gpioMap.put(gpio.getGpioName(), gpio);
				gpioList.put(gpio, new GPIOValueList());
			}
		} catch (Exception e) {

		} finally {
			lock.unlock();
		}
	}

	public IOGPIO getGpio(String name) {
		return gpioMap.get(name);
	}

	public void addGpioStateHandler(IOGpioState stateHandler) {
		gpioStateList.add(stateHandler);
	}

	public void removeGpioStateHandler(IOGpioState stateHandler) {
		gpioStateList.remove(stateHandler);
	}

	public void unRegister(IOGPIO gpio) {
		try {
			lock.lock();
			if (gpioList.containsKey(gpio)) {
				gpioMap.remove(gpio.getGpioName());
				gpioList.remove(gpio);
			}
		} catch (Exception e) {

		} finally {
			lock.unlock();
		}
	}

	@Override
	protected boolean IOOpen() throws IOException {
		opened = true;
		thread = new GPIOThread();
		thread.start();
		return opened;
	}

	@Override
	protected boolean IOClose() throws IOException {
		opened = false;
		try {
			thread.join();
		} catch (InterruptedException e) {

		}
		try {
			lock.lock();
			for (IOGPIO gpio : gpioList.keySet()) {
				try {
					gpio.close();
				} catch (Exception e) {

				}

			}
		} catch (Exception e) {

		} finally {
			lock.unlock();
		}
		return !opened;
	}

	class GPIOThread extends Thread {
		@Override
		public void run() {
			this.setName(THREAD_NAME);
			int inputCount = 0;
			while (isOpened()) {
				try {
					lock.lock();
					for (IOGPIO gpio : gpioList.keySet()) {
						if (!(gpio instanceof IOGpioInput)) {
							continue;
						}
						inputCount++;
						boolean oldValue = gpioList.get(gpio).get();
						boolean value = ((IOGpioInput) gpio).get();
						gpioList.get(gpio).add(value);
						value = gpioList.get(gpio).get();
						for (IOGpioState stateHandler : gpioStateList) {
							stateHandler.onGpioState(((IOGpioInput) gpio), value);
							if (oldValue != value) {
								stateHandler.onGpioStateChange(((IOGpioInput) gpio), value);
							}

						}
					}
				} catch (Exception e) {
					e.printStackTrace();
				} finally {
					lock.unlock();
					if (inputCount == 0) {
						return;
					}
					inputCount = 0;
					try {
						while (paused && isOpened()) {
							Thread.sleep(1000);
						}
						Thread.sleep(requestTimeOut);
					} catch (InterruptedException e) {

					}
				}
			}

		}
	}
}
